@@ -36,6 +36,7 @@ from Spanners.Squirrel import Squirrel
 from Spanners.Treeify import Treeify
 from Baubles.Logger import Logger
 from Baubles.Colours import Colours
+from GoldenChild.xpath import *
 
 from sqlalchemy.engine import reflection
 from sqlalchemy import not_
@@ -134,6 +135,9 @@ class Empty(object):
 		
 @args.command(single=True)
 class Export(object):
+	'''
+	tool to inspect, query and edit Knowledge Base sqlite files
+	'''
 	
 	@args.property(short='v', flag=True, help='verbose logging')
 	def verbose(self):
@@ -269,6 +273,10 @@ class Export(object):
 	@args.operation
 	@args.parameter(name='klobber', short='k', flag=True, help='create from scratch')
 	def load_opml(self, file, klobber=False):
+		'''
+		load OPML xml format into the KDB
+		'''
+		
 		if klobber:
 			Empty().toDB(Empty.base64, self.database) # create clean copy from empty.base64
 
@@ -288,6 +296,10 @@ class Export(object):
 		
 	@args.operation
 	def load_excel(self, file):
+		'''
+		load an excel sheet into the KDB file
+		'''
+		
 		session = self.Session()
 		
 		school = session.query(Item).filter_by(Name="FSHS").first()
@@ -352,72 +364,84 @@ class Export(object):
 		session.commit()
 		session.close()			
 		
+
+	def _import_node(self, session, cod, childItem, parent, indent=''):
+		'''
+		load an item and iterate through children
+		'''
+
+		text = getElementText(cod.ctx, 'title', childItem)
+		print(f'{indent}{text}')
+
+		notes = getElement(cod.ctx, 'notes', childItem)
+		note = None
+		
+		if notes:
+			note = notes.content
+			if note != '(null)':
+				print(f'{indent}{indent}"{note}"')
+				
+		item = Item()
+		session.add(item)
+		item.Name = text
+		item.Description = note
+
+		relation = Relation()
+		session.add(relation)
+		#relation.Name = 'r1'
+		#relation.Description = 'Relation One'
+		relation.inbound = parent
+		relation.outbound = item
+
+		for grandChild in getElements(cod.ctx, 'ChildItem', childItem):
+			self._import_node(session, cod, grandChild, item, indent=f'{indent}  ')
+			
 		
 	@args.operation(name="import")
-	def importer(self):
+	@args.parameter(name='name', help='name of node to append to')
+	@args.parameter(name='file', help='cloud outliner cod file')
+	def importer(self, name, file):
+		'''
+		import cloud outliner cod file to a node parent
+		'''
+		
 		session = self.Session()
-		
-		item1 = Item()
-		session.add(item1)
-		item1.Name = 'i1'
-		item1.Description = 'Item One'
-		
-		item2 = Item()
-		session.add(item2)
-		item2.Name = 'i2'
-		item2.Description = 'Item Two'
 
-		relation1 = Relation()
-		session.add(relation1)
-		relation1.Name = 'r1'
-		relation1.Description = 'Relation One'
-		relation1.inbound = item1
-		relation1.outbound = item2
+		root = session.query(Item).filter_by(Name=name).first()
+
+		if not root:
+			root = Item()
+			session.add(root)
+			root.Name = file
+
+		cod = XML(*getContextFromFile(os.path.expanduser(file)))
+		
+		for childItem in getElements(cod.ctx, '/Document/Properties/context/ChildItem'):
+			self._import_node(session, cod, childItem, root, indent='')
 		
 		session.commit()
-
-		print(relation1.IDTo, relation1.IDFrom)
-		
 		session.close()
 
 
 	@args.operation
-	@args.parameter(name='name', short='n', help='the name of the element')
-	@args.parameter(name='references', short='r', flag=True, help='list references')
-	@args.parameter(name='categories', short='c', flag=True, help='list categories')
-	@args.parameter(name='attachments', short='a', flag=True, help='list attachments')
-	def inspect(self, name=None, references=False, categories=False, attachments=False):
+	@args.parameter(name='references', short='r', flag=True, help='get the reference categories instead')
+	def categories(self, references=False):
+		'''
+		inspect the structure for categories for the templates in the KDB
+		'''
+		
 		session = self.Session()
 
 		if references:
-			query = session.query(Relation)
-			for reference in query.all():
-				print(f'{reference.inbound.Name} -> {reference.outbound.Name} : {reference.Name}')
-			return
-		
-		if categories:
+			query = session.query(RelationTemplate)
+			for relation in query.all():
+				print(relation.TemplateName)
+					  
+		else:
 			query = session.query(ItemTemplate)
-			if name: query = query.filter_by(Name=name)
 			for category in query.all():
-				print(category.ItemTemplateID, category.TemplateName)
-			return
+				print(category.TemplateName)
 		
-		if attachments:
-			query = session.query(ItemAttachment)
-			if name: query = query.join(ItemAttachment.item).filter(Item.Name == name)
-			for attachment in query.all():
-				print(attachment.item.Name[:20], attachment.attachment.Data[:60])
-			return
-		
-		query = session.query(Item)
-		if name: query = query.filter_by(Name=name)
-
-		for item in query.all(): 
-			print(item.Name)
-			if references:
-				for reference in item.outbound:
-					print('\t%s'%reference.outbound.Name)
-
 		session.close()
 
 
@@ -434,12 +458,63 @@ class Export(object):
 				sequence += 10
 				self._sort(session, relation.outbound, depth-1, indent='%s\t'%indent)
 
-				
+
+	@args.operation
+	def categorise(self, asis, tobe):
+		'''
+		change the category from asis to tobe
+		empty string "" for None
+		'''
+
+		session = self.Session()
+
+		if asis == '':
+			_asis = None
+			print('from: <None>')
+		else:
+			_asis = session.query(ItemTemplate).filter_by(TemplateName=asis).first()
+			if _asis:
+				print(f'from: {_asis.TemplateName}')
+			else:
+				sys.stderr.write(f'can\' find category {asis}\n')
+				return
+
+		if tobe == '':
+			_tobe = None
+			print('to: <None>')
+		else:
+			_tobe = session.query(ItemTemplate).filter_by(TemplateName=tobe).first()
+			if _tobe:
+				print(f'to:  -> {_tobe.TemplateName}')
+			else:
+				sys.stderr.write(f'can\'t find category {tobe}\n')
+				return
+
+		if _asis:
+			query = session.query(Item).filter_by(ItemTemplateID=_asis.ItemTemplateID)
+		else:
+			query = session.query(Item).filter_by(ItemTemplateID=-1)
+			
+		for item in query.all():
+			print(f'{item.Name} : <{item.ItemTemplateID}> ')
+			if _tobe:
+				item.category = _tobe
+			else:
+				item.ItemTemplateID = -1
+		
+		session.commit()
+		session.close()
+
+		
 	@args.operation
 	@args.parameter(name='name', help='name of node to sort references for')
 	@args.parameter(name='depth', short='d', type=int, help='recursion depth')
 	@args.parameter(name='noCommit', short='n', flag=True, help='no commit for changes')
 	def sort(self, name, depth=1, noCommit=False):
+		'''
+		sort an items child items
+		'''
+		
 		session = self.Session()
 
 		for item in session.query(Item).filter_by(Name=name):
@@ -449,29 +524,83 @@ class Export(object):
 			session.commit()
 		session.close()
 
-		
+
 	@args.operation
-	@args.parameter(name='value', help='string to search in name for')
+	@args.parameter(name='name', help='string to search in name for')
+	@args.parameter(name='description', short='d', flag=True, help='include description')
+	@args.parameter(name='url', short='u', flag=True, help='include url')
+	@args.parameter(name='attachments', short='a', flag=True, help='include attachments')
+	@args.parameter(name='references', short='r', choices={'i':'inbound','o':'outbound','b':'both'}, help='include references')
 	@args.parameter(name='categories', short='c', nargs='+',  help='restrict to categories')
-	def query(self, value, description=False, categories=[]):
+	def query(self, name, description=None, url=None, attachments=None, references=None, categories=[]):
+		'''
+		query a value, and filter by category
+		'''
+		
 		session = self.Session()
 		
-		query = session.query(Item).filter(Item.Name.like(value)).join(Item.category)
+		query = session.query(Item).filter(Item.Name.like(name)).join(Item.category)
 
 		for item in query.all():
 			if categories:
 				if item.category.TemplateName not in categories:
 					continue
-			print(item.Name)
+
+			text = f'{item.Name}'
+			if item.category:
+				text = f'{text} : <{item.category.TemplateName}>'
+			print(text)
+			
+			if description:
+				print(f'\t# {item.Description}')
+				
+			if url:
+				print(f'\t@ {item.DirectLinkURL}')
+				
+			if attachments:
+				for attachment in item.attachments:
+					print(f'\t& {attachment.attachment.Data}')
+					
+			if references in ['i','b']:
+				for reference in item.inbound:
+					text = f'\t< {reference.inbound.Name}'
+					if reference.category:
+						text = f'{text} : <{reference.category.TemplateName}>'
+					print(text)
+					
+			if references in ['o','b']:
+				for reference in item.outbound:
+					text = f'\t> {reference.outbound.Name}'
+					if reference.category:
+						text = f'{text} : <{reference.category.TemplateName}>'
+					print(text)
 				
 		session.close()
 
+
+	@args.operation
+	def step_in(self):
+		'''
+		todo
+		'''
+
+
+	@args.operation
+	def step_out(self):
+		'''
+		todo
+		'''
 		
+
 	@args.operation
 	@args.parameter(name='old', help='string to search in name for')
 	@args.parameter(name='new', help='string to replace with')
 	@args.parameter(name='description', short='d', flag=True, help='search in description as well')
 	def replace(self, old, new, description=False):
+		'''
+		search and replace in item names
+		'''
+		
 		session = self.Session()
 		for item in session.query(Item):
 			_name = item.Name
@@ -493,6 +622,10 @@ class Export(object):
 		
 	@args.operation
 	def clean(self):
+		'''
+		remote html tags from descriptions
+		'''
+		
 		session = self.Session()
 		for item in session.query(Item):
 			_name = item.Name
@@ -509,7 +642,11 @@ class Export(object):
 
 
 	@args.operation
-	def export(self):
+	def to_csv(self):
+		'''
+		export data to directory as tables of raw KDB types
+		'''
+		
 		shutil.copy2(self.database, '%s/.kdb'%self.exportdir)
 		self.inspector =  sqlalchemy.inspect(self.__engine)
 		for table in self.inspector.get_table_names():
@@ -520,6 +657,10 @@ class Export(object):
 
 	@args.operation
 	def deArrow(self):
+		'''
+		remove -> arrows from descriptions
+		'''
+		
 		session = self.Session()
 
 		for item in session.query(Item).filter(Item.ItemID>0): 
